from typing import Tuple
from sympy import mod_inverse, isprime
from functools import lru_cache

# Base must be co-prime to HASH_MODULUS
# Since HASH_MODULUS is a prime:
# Base must not be a multiple of HASH_MODULUS

HASH_MODULUS = 18446744073709551557 # largest prime smaller than 2^64
HASH_INITIAL = 0x0

def rolling_hash(s: bytes, base=255):
    sum = HASH_INITIAL
    for i in range(len(s)):
        sum = (sum * base) % HASH_MODULUS
        sum = (sum + s[i]) % HASH_MODULUS

    return normalize_hash(sum), len(s)

@lru_cache(maxsize=None)
def mult_inv(num, base=255):
    return (mod_inverse(base, HASH_MODULUS) ** num) % HASH_MODULUS

def normalize_hash(h):
    if h >= 0:
        return h
    h = h + ((-(h // HASH_MODULUS) + 1) * HASH_MODULUS)
    return h

def rolling_hash_drop_prefix(rhash: Tuple[int, int], s: bytes, base=255):
    h, n = rhash
    new_hash, m = rolling_hash(s, base)

    h = h - (new_hash * (base**(n-m)) % HASH_MODULUS) % HASH_MODULUS
    return normalize_hash(h), n - m

def rolling_hash_drop_suffix(rhash: Tuple[int, int], s: bytes, base=255):
    h, n = rhash
    new_hash, m = rolling_hash(s, base)

    h = (h - new_hash) * mult_inv(m, base) % HASH_MODULUS
    return normalize_hash(h), n - m

def rolling_hash_append(rhash: Tuple[int, int], s: bytes, base=255):
    sum, n = rhash
    for i in range(len(s)):
        sum = (sum * base) % HASH_MODULUS
        sum = (sum + s[i]) % HASH_MODULUS
    return normalize_hash(sum), n + len(s)
