import hash

class RepetitionNode:
    def __init__(self, content, N=1):
        self.content = content
        self.N = N

    def _expand(self):
        if isinstance(self.content, bytes):
            return self.content * self.N
        elif isinstance(self.content, list):
            return expand(self.content * self.N)
        elif isinstance(self.content, RepetitionNode):
            return expand(self.content) * self.N

    def __repr__(self):
        if self.N == 1:
            if isinstance(self.content, bytes) and len(self.content) > 1:
                return f'({self.content.decode()})'
            elif isinstance(self.content, list):
                return f'({" ".join(rep_to_str(x) for x in self.content)})'
            else:
                return f'{self.content}'
        else:
            if isinstance(self.content, bytes):
                if len(self.content) > 1:
                    return f'({self.content.decode()})^{self.N}'
                else:
                    return f'{self.content.decode()}^{self.N}'
            elif isinstance(self.content, list):
                return f'({" ".join(rep_to_str(x) for x in self.content)})^{self.N}'
            else:
                return f'({self.content})^{self.N}'

    def _flatten(self):
        if isinstance(self.content, list):
            while isinstance(self.content, list) and len(self.content) == 1:
                self.content = self.content[0]
            self.content = flatten(self.content)
        if isinstance(self.content, RepetitionNode):
            self.content._flatten()
            self.N = self.content.N * self.N
            self.content = self.content.content
        return self

def flatten(x):
    if isinstance(x, RepetitionNode):
        x = x._flatten()
        while isinstance(x, RepetitionNode) and x.N == 1:
            x = flatten(x.content)
        return x
    elif isinstance(x, list):
        x = [flatten(i) for i in x]
        ret = []
        for i in x:
            if isinstance(i, list):
                ret += flatten(i)
            else:
                ret.append(i)
        return ret
    else:
        return x

def expand(x):
    if isinstance(x, list):
        return b''.join([expand(i) for i in x])
    elif isinstance(x, bytes):
        return x
    elif isinstance(x, RepetitionNode):
        return x._expand()

def rep_to_str(x):
    if isinstance(x, list):
        return ' '.join([rep_to_str(i) for i in x])
    elif isinstance(x, bytes):
        return x.decode()
    elif isinstance(x, RepetitionNode):
        return str(x)

def detect_repetitions(s: bytes):
    n = len(s)
    if n < 1:
        return None

    if n == 1:
        return s

    structure = []
    abs_structure = []

    # TODO: Bei rekursiven aufrufen den hash übergeben

    last = hash.rolling_hash(s[0:(len(s)+1)//2+1])
    prev_hashes =[last]

    for i in range(len(s)//2-1):
        last = hash.rolling_hash_drop_prefix(last, s[i:i+1])
        last = hash.rolling_hash_append(last, s[i+(len(s)+1)//2+1:i+(len(s)+1)//2+2])
        prev_hashes.append(last)


    for i in reversed(range(((len(s)+1)//2)+1)):
        more_hashes = [
            hash.rolling_hash_drop_suffix(
                prev_hashes[j], s[j+i:j+i+1]
            ) for j in range(len(prev_hashes))
        ]
        last_hash   = [
            hash.rolling_hash_drop_prefix(prev_hashes[-1], s[-i-1:-i])
        ]

        length_hashes = i

        # more_hashes = [hash.hash_roll_poly2(s[j:j+length_hashes]) for j in range(len(s)-i-1)]

        prev_hashes = more_hashes + last_hash

        # print(f"{prev_hashes=}")

        # print(prev_hashes)
        # print(len(prev_hashes))

        sequences = [[]] * n
        avail = []

        for offset in range(length_hashes):
            last = None
            rep = 0
            start = 0

            for elem_idx in range(len(prev_hashes)):
                idx = elem_idx * length_hashes + offset
                if idx >= len(prev_hashes):
                    break

                if last == prev_hashes[idx]:
                    # print(f'{s[start:start+length_hashes]=} {s[idx:idx+length_hashes]=}')
                    if s[start:start+length_hashes] == s[idx:idx+length_hashes]:
                        # print('wow')
                        rep += 1
                        continue
                    else:
                        # TODO: Kollisionen zählen
                        # pass
                        print('collision! zomg')
                if rep > 0:
                    sequences[start].append((start,rep))
                    avail.append(start)
                    rep = 0
                    # start = start + rep * length_hashes
                start = idx
                last = prev_hashes[idx]
            if rep > 0:
                sequences[start].append((start,rep))
                avail.append(start)
                rep = 0

        if len(avail) < 1:
            continue

        last = (-1,-1)
        for x in avail:
            elem = sequences[x][0] # XXX is there only one element per sequences[x]
            if elem[0] <= (last[0] + last[1] * length_hashes):
                continue
            last = elem
            structure.append(elem)

        if len(structure) != 0:
            if structure[0][0] != 0:
                abs_structure.append(detect_repetitions(s[0:structure[0][0]]))
            last_start = None
            for start, rep in structure:
                if last_start is not None and last_start + (last_rep + 1) * length_hashes != start:
                    abs_structure.append(detect_repetitions(s[last_start + (last_rep + 1) * length_hashes:start]))
                abs_structure.append(RepetitionNode(detect_repetitions(s[start:start + length_hashes]), rep+1))
                last_start = start
                last_rep = rep
            if structure[-1][0] + ((structure[-1][1]+1) * length_hashes) < len(s):
                abs_structure.append(detect_repetitions(s[structure[-1][0]+(structure[-1][1]+1) * length_hashes:]))
            return abs_structure
    return RepetitionNode(s)
