// Copyright (C) 2025 Johannes Knödtel
// Full notice in 'LICENSE'

#include <stdint.h>
#include "multinv.h"
#include "hash.h"

/* typedef _BitInt(128) int128_t */
typedef __int128 int128_t;

uint32_t power_of_base(int pow)
{
    uint64_t result = 1;
    for (int i = 1; i <= pow; i++)
    {
        result = result * BASE;
        result = result % PRIME;
    }
    return (uint32_t)result;
}

static uint64_t normalize_hash(uint64_t x)
{
    while (x < 0)
    {
        x += PRIME;
    }

    return x;
}

void hash_roll_poly2(int length, char signs[], hashDesc *results)
{
    uint64_t sum = HASH_INITIAL;
    for (int i = 0; i < length; i++)
    {
        sum = ((uint64_t)sum * BASE) % PRIME;
        sum = ((uint64_t)sum + signs[i]) % PRIME;
    }
    results->h = (uint32_t)sum;
    results->len = length;
}

void hash_drop_prefix_poly2(hashDesc *hashArr, char signs[], int index, hashDesc *results)
{
    hashDesc resultPrefix;
    char signsIntern[1];
    signsIntern[0] = signs[index];
    hash_roll_poly2(1, signsIntern, &resultPrefix);

    uint32_t hash =
        normalize_hash((hashArr->len - ((uint64_t)resultPrefix.h * (power_of_base(hashArr->len - 1) % PRIME)) % PRIME));

    results->h = hash;
    results->len = hashArr->len - 1;
}

void hash_drop_suffix_poly2(hashDesc *hashArr, char signs[], int index, hashDesc *results)
{
    char signsIntern[1];
    signsIntern[0] = signs[index + (hashArr->len - 1)];
    uint32_t test = multiplicative_inverses[BASE];

    uint32_t hash = normalize_hash((((uint64_t)hashArr->h - signsIntern[0]) * test) % PRIME);

    results->h = hash;
    results->len = hashArr->len - 1;
}
