// Copyright (C) 2025 Johannes Knödtel
// Full notice in 'LICENSE'

#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <assert.h>

#include <repetition.h>
#include <hash.h>

#define NDEBUG

repetitionNode *initRepNode(int n)
{
    repetitionNode *a = malloc(sizeof(repetitionNode));
    a->N = n + 1;
    a->lengthOfNode = 0;
    a->firstBase = NULL;
    a->firstRep = NULL;
    a->lastBase = NULL;
    a->lastRep = NULL;
    a->nextBase = NULL;
    a->nextRep = NULL;

    return a;
}

static baseNode *initBaseNode(char characterHere)
{
    baseNode *a = malloc(sizeof(baseNode));
    a->character = characterHere;
    a->nextBase = NULL;
    a->nextRep = NULL;

    return a;
}

static void flatten(repetitionNode *node)
{
    if (node->firstRep != NULL && node->firstRep == node->lastRep)
    {
        node->N = node->N * node->firstRep->N;

        if (node->firstRep->firstBase == NULL)
        {
            node->firstRep = node->firstRep->firstRep;
        }
        else
        {
            node->firstBase = node->firstRep->firstBase;
            node->firstRep = NULL;
        }

        if (node->lastRep->lastBase == NULL)
        {
            node->lastRep = node->lastRep->lastRep;
        }
        else
        {
            node->lastBase = node->lastRep->lastBase;
            node->lastRep = NULL;
        }
        flatten(node);
    }
    else
    {
        baseNode *tempBase = node->firstBase;
        repetitionNode *tempRep = node->firstRep;
        while (tempRep != NULL || tempBase != NULL)
        {
            if (tempRep != NULL)
            {
                flatten(tempRep);
                if (tempRep->nextBase == NULL)
                {
                    tempRep = tempRep->nextRep;
                }
                else
                {
                    tempBase = tempRep->nextBase;
                    tempRep = NULL;
                }
            }
            else
            {
                if (tempBase->nextBase == NULL)
                {
                    tempRep = tempBase->nextRep;
                    tempBase = NULL;
                }
                else
                {
                    tempBase = tempBase->nextBase;
                }
            }
        }
    }
}

static void appendToRepNode(repetitionNode *list, repetitionNode *rep, baseNode *base)
{
    if (rep == NULL && base == NULL)
    {
        printf("No Node to append!");
    }
    else if (rep == NULL)
    {
        if (list->firstBase == NULL && list->firstRep == NULL)
        {
            list->firstBase = base;
            list->lastBase = base;
        }
        else if (list->lastBase == NULL)
        {
            list->lastRep->nextBase = base;
            list->lastBase = base;
            list->lastRep = NULL;
        }
        else if (list->lastRep == NULL)
        {
            list->lastBase->nextBase = base;
            list->lastBase = base;
        }
    }
    else
    {
        if (list->firstBase == NULL && list->firstRep == NULL)
        {
            list->firstRep = rep;
            list->lastRep = rep;
        }
        else if (list->lastBase == NULL)
        {
            list->lastRep->nextRep = rep;
            list->lastRep = rep;
        }
        else if (list->lastRep == NULL)
        {
            list->lastBase->nextRep = rep;
            list->lastRep = rep;
            list->lastBase = NULL;
        }
    }
}

void printRepNode(repetitionNode *a)
{
    printf("(");
    if (a->firstBase == a->lastBase && a->firstBase != NULL)
    {
        printf("%c", a->firstBase->character);
    }
    else
    {
        baseNode *tempBase = a->firstBase;
        repetitionNode *tempRep = a->firstRep;
        while (1)
        {
            if (tempBase == a->lastBase && tempBase != NULL)
            {
                printf("%c", tempBase->character);
                break;
            }
            else if (tempRep == a->lastRep && tempRep != NULL)
            {
                printRepNode(tempRep);
                break;
            }
            else if (tempBase == NULL)
            {
                printRepNode(tempRep);
                if (tempRep->nextBase == NULL)
                {
                    tempRep = tempRep->nextRep;
                }
                else
                {
                    tempBase = tempRep->nextBase;
                    tempRep = NULL;
                }
            }
            else
            {
                printf("%c", tempBase->character);
                if (tempBase->nextBase == NULL)
                {
                    tempRep = tempBase->nextRep;
                    tempBase = NULL;
                }
                else
                {
                    tempBase = tempBase->nextBase;
                }
            }
        }
    }
    printf(")^%d ", a->N);
}

static int detect_repetitions_intern(repetitionNode *origNode, char signs[], int length, int currentAmountOfHashes, hashDesc *hashes)
{
    if (length < 1)
    {
        free(hashes);
    }
    else if (length == 1)
    {
        free(hashes);
        baseNode *node = initBaseNode(signs[0]);
        appendToRepNode(origNode, NULL, node);
    }
    else
    {
        int FoundRepetition = 0;
        for (int hashLength = hashes[0].len; hashLength >= 1 && FoundRepetition == 0; hashLength--)
        {
            int repetitionAmount = 0;
            int index = 0;
            int startIndex = index;
            int firstRepFound = 1;
            for (; index < currentAmountOfHashes;)
            {
                if (index + hashLength >= currentAmountOfHashes)
                {
                    if (repetitionAmount > 0)
                    {
                        repetitionNode *repNode = initRepNode(repetitionAmount);
                        int tempSizeHashInNode = (hashLength + 1) / 2 + 1; // +1 is there to round up
                        hashDesc *hashesInNode = malloc(sizeof(hashDesc) * tempSizeHashInNode);
                        assert(hashesInNode);

#ifdef _OPENMP
                        if (LIMIT_OMP > (hashLength + 1) / 2 + 1)
                        {
                            for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                            {
                                hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                            }
                        }
                        else
                        {
#pragma omp parallel for
                            for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                            {
                                hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                            }
                        }
#else
                        for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                        {
                            hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                        }
#endif

                        detect_repetitions_intern(repNode, signs + index, hashLength, tempSizeHashInNode, hashesInNode);
                        appendToRepNode(origNode, repNode, NULL);
                        startIndex = index + hashLength;
                        index = index + hashLength;
                        firstRepFound = 1;
                        repetitionAmount = 0;
                    }
                    index++;
                    break;
                }
                if (hashes[index].h == hashes[index + hashLength].h)
                {
                    int test = 1;
                    for (int h = index; h < index + hashLength; h++)
                    {
                        if (signs[h] != signs[h + hashLength])
                            test = 0;
                    }
                    if (test)
                    {
                        if (firstRepFound)
                        {
                            int sizeOfSignsBefore = index - startIndex;

                            if (sizeOfSignsBefore > 0)
                            {
                                int tempSizeOfHashesBeforeNode = (sizeOfSignsBefore + 1) / 2 + 1; // +1 is there to round up
                                hashDesc *hashesBeforeNode = malloc(sizeof(hashDesc) * tempSizeOfHashesBeforeNode);
                                assert(hashesBeforeNode);

#ifdef _OPENMP
                                if (LIMIT_OMP > (sizeOfSignsBefore + 1) / 2 + 1)
                                {
                                    for (int i = 0; i < (sizeOfSignsBefore + 1) / 2 + 1; i++)
                                    {
                                        hash_roll_poly2(sizeOfSignsBefore / 2, signs + startIndex + i, hashesBeforeNode + i);
                                    }
                                }
                                else
                                {
#pragma omp parallel for
                                    for (int i = 0; i < (sizeOfSignsBefore + 1) / 2 + 1; i++)
                                    {
                                        hash_roll_poly2(sizeOfSignsBefore / 2, signs + startIndex + i, hashesBeforeNode + i);
                                    }
                                }
#else
                                for (int i = 0; i < (sizeOfSignsBefore + 1) / 2 + 1; i++)
                                {
                                    hash_roll_poly2(sizeOfSignsBefore / 2, signs + startIndex + i, hashesBeforeNode + i);
                                }
#endif

                                detect_repetitions_intern(origNode, signs + startIndex, sizeOfSignsBefore, tempSizeOfHashesBeforeNode, hashesBeforeNode);
                            }
                            firstRepFound = 0;
                        }
                        FoundRepetition = 1;
                        repetitionAmount++;
                        index = index + hashLength;
                    }
                    else
                    {
                        if (repetitionAmount > 0)
                        {
                            repetitionNode *repNode = initRepNode(repetitionAmount);
                            int tempSizeHashInNode = (hashLength + 1) / 2 + 1; // +1 is there to round up
                            hashDesc *hashesInNode = malloc(sizeof(hashDesc) * tempSizeHashInNode);
                            assert(hashesInNode);

#ifdef _OPENMP
                            if (LIMIT_OMP > (hashLength + 1) / 2 + 1)
                            {
                                for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                                {
                                    hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                                }
                            }
                            else
                            {
#pragma omp parallel for
                                for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                                {
                                    hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                                }
                            }
#else
                            for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                            {
                                hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                            }
#endif

                            detect_repetitions_intern(repNode, signs + index, hashLength, tempSizeHashInNode, hashesInNode);
                            appendToRepNode(origNode, repNode, NULL);
                            startIndex = index + hashLength;
                            index = index + hashLength;
                            firstRepFound = 1;
                            repetitionAmount = 0;
                        }
                        index++;
                    }
                }
                else
                {
                    if (repetitionAmount > 0)
                    {
                        repetitionNode *repNode = initRepNode(repetitionAmount);
                        int tempSizeHashInNode = (hashLength + 1) / 2 + 1; // +1 is there to round up
                        hashDesc *hashesInNode = malloc(sizeof(hashDesc) * tempSizeHashInNode);
                        assert(hashesInNode);

#ifdef _OPENMP
                        if (LIMIT_OMP > (hashLength + 1) / 2 + 1)
                        {
                            for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                            {
                                hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                            }
                        }
                        else
                        {
#pragma omp parallel for
                            for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                            {
                                hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                            }
                        }
#else
                        for (int i = 0; i < (hashLength + 1) / 2 + 1; i++)
                        {
                            hash_roll_poly2(hashLength / 2, signs + index + i, hashesInNode + i);
                        }
#endif

                        detect_repetitions_intern(repNode, signs + index, hashLength, tempSizeHashInNode, hashesInNode);
                        appendToRepNode(origNode, repNode, NULL);
                        startIndex = index + hashLength;
                        index = index + hashLength;
                        firstRepFound = 1;
                        repetitionAmount = 0;
                    }
                    else
                    {
                        index++;
                    }
                }
            }
            if (startIndex < length && FoundRepetition)
            {
                int sizeOfSignsAfter = length - startIndex;
                int tempSizePost = (sizeOfSignsAfter + 1) / 2 + 1;
                hashDesc *hashesPostNode = malloc(sizeof(hashDesc) * tempSizePost); // +1 is there to round up
                assert(hashesPostNode);

#ifdef _OPENMP
                if (LIMIT_OMP > (sizeOfSignsAfter + 1) / 2 + 1)
                {
                    for (int i = 0; i < (sizeOfSignsAfter + 1) / 2 + 1; i++)
                    {
                        hash_roll_poly2(sizeOfSignsAfter / 2, signs + startIndex + i, hashesPostNode + i);
                    }
                }
                else
                {
#pragma omp parallel for
                    for (int i = 0; i < (sizeOfSignsAfter + 1) / 2 + 1; i++)
                    {
                        hash_roll_poly2(sizeOfSignsAfter / 2, signs + startIndex + i, hashesPostNode + i);
                    }
                }
#else
                for (int i = 0; i < (sizeOfSignsAfter + 1) / 2 + 1; i++)
                {
                    hash_roll_poly2(sizeOfSignsAfter / 2, signs + startIndex + i, hashesPostNode + i);
                }
#endif

                free(hashes);
                detect_repetitions_intern(origNode, signs + startIndex, sizeOfSignsAfter, tempSizePost, hashesPostNode);
            }

            if (FoundRepetition == 0 && hashLength > 1)
            {
                hashDesc *newHashes = malloc(sizeof(hashDesc) * (currentAmountOfHashes + 1));
                assert(newHashes);

#ifdef _OPENMP
                if (LIMIT_OMP > currentAmountOfHashes)
                {
                    for (int i = 0; i < currentAmountOfHashes; i++)
                    {
                        hash_drop_suffix_poly2(hashes + i, signs, i, newHashes + i);
                    }
                }
                else
                {
#pragma omp parallel for
                    for (int i = 0; i < currentAmountOfHashes; i++)
                    {
                        hash_drop_suffix_poly2(hashes + i, signs, i, newHashes + i);
                    }
                }
#else
                for (int i = 0; i < currentAmountOfHashes; i++)
                {
                    hash_drop_suffix_poly2(hashes + i, signs, i, newHashes + i);
                }
#endif
                hash_drop_prefix_poly2(hashes + (currentAmountOfHashes - 1), signs,
                                       currentAmountOfHashes - 1, newHashes + currentAmountOfHashes);
                free(hashes);
                detect_repetitions_intern(origNode, signs, length, currentAmountOfHashes + 1, newHashes);
                return 0;
            }
        }
        if (FoundRepetition == 0)
        {
            for (int i = 0; i < length; i++)
            {
                baseNode *node = initBaseNode(signs[i]);
                appendToRepNode(origNode, NULL, node);
            }
            free(hashes);
        }
    }
    return 0;
}

void detect_repetitions(repetitionNode *origNode, char signs[], int length)
{
    int tempLengthOfHashes = (length + 1) / 2 + 1;
    hashDesc *hashes = malloc(sizeof(hashDesc) * tempLengthOfHashes); // +1 is there to round up
    assert(hashes);

#ifdef _OPENMP
    if (LIMIT_OMP > (length + 1) / 2 + 1)
    {
        for (int i = 0; i < (length + 1) / 2 + 1; i++)
        {
            int tempSizeSigns = length / 2;
            hash_roll_poly2(length / 2, signs + i, hashes + i);
        }
    }
    else
    {
#pragma omp parallel for
        for (int i = 0; i < (length + 1) / 2 + 1; i++)
        {
            int tempSizeSigns = length / 2;
            hash_roll_poly2(length / 2, signs + i, hashes + i);
        }
    }
#else
    for (int i = 0; i < (length + 1) / 2 + 1; i++)
    {
        int tempSizeSigns = length / 2;
        hash_roll_poly2(length / 2, signs + i, hashes + i);
    }
#endif

    int currentSizeOfHashes = tempLengthOfHashes;
    detect_repetitions_intern(origNode, signs, length, currentSizeOfHashes, hashes);
    flatten(origNode);
}